#define BOOST_AUTO_TEST_MAIN
#include <boost/test/auto_unit_test.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
namespace fs=boost::filesystem;

#include <boost/algorithm/string/case_conv.hpp>

#include <list>
#include <iostream>
#include <sstream>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>

#include "alg/sequence.hpp"
#include "mussa_exceptions.hpp"

using namespace std;

BOOST_AUTO_TEST_CASE( sequence_copy_constructor )
{
  SequenceRef s(new Sequence("AAAAGGGG"));
  s->set_species("foo");
  BOOST_CHECK_EQUAL(s->get_species(), "foo");
  
  SequenceRef c(new Sequence(s));
  BOOST_CHECK_EQUAL(c->get_species(), "foo");

  c->set_species("bar");
  BOOST_CHECK_EQUAL(s->get_species(), "foo");
  BOOST_CHECK_EQUAL(c->get_species(), "bar");
}
BOOST_AUTO_TEST_CASE( sequence_get_sequence )
{
	Sequence s;
	// make sure that retrieving the sequence doesn't throw an error
	BOOST_CHECK_EQUAL(s.get_sequence(), std::string() );
}

BOOST_AUTO_TEST_CASE( sequence_from_string )
{
  std::string str1("AAAT");
  Sequence seq1(str1);
  BOOST_CHECK_EQUAL(seq1.get_sequence(), str1);
} 

BOOST_AUTO_TEST_CASE( sequence_find_first_not_of )
{
  std::string str1("AAAAT");
  Sequence seq1(str1);
  BOOST_CHECK_EQUAL(seq1.find_first_not_of("A"), str1.find_first_not_of("A"));
  
  std::string str2("AATTGGCC");
  Sequence seq2(str2);
  BOOST_CHECK_EQUAL(seq2.find_first_not_of("qwer"), str2.find_first_not_of("qwer"));
}

//! when we try to load a missing file, do we get an error?
BOOST_AUTO_TEST_CASE( sequence_load_exception )
{
  Sequence s;
  // there should be errors when we try to load something that doesn't exist
  BOOST_CHECK_THROW( s.load_fasta("alkejralk", 1, 0, 0), mussa_load_error);
  BOOST_CHECK_THROW( s.load_annot("alkejralk", 0, 0), mussa_load_error);
}

BOOST_AUTO_TEST_CASE( sequence_eol_conventions )
{
  string header(">Header");
  string line1("AAAAGGGGCCCCTTTTT");
  string line2("AAAAGGGGCCCCTTTTT");
  int seq_len = line1.size() + line2.size();

  stringstream cr;
  cr << header << "\015" << line1 << "\015" << line2 << "\015";
  Sequence seq_cr;
  seq_cr.load_fasta(cr);

  stringstream crlf;
  crlf << header << "\015\012" << line1 << "\015\012" << line2 << "\015\012";
  Sequence seq_crlf;
  seq_crlf.load_fasta(crlf);

  stringstream lf;
  lf << header << "\012" << line1 << "\012" << line2 << "\012";
  Sequence seq_lf;
  seq_lf.load_fasta(lf);

  BOOST_CHECK_EQUAL(seq_cr.size(),   seq_len);
  BOOST_CHECK_EQUAL(seq_crlf.size(), seq_len);
  BOOST_CHECK_EQUAL(seq_cr.size(),   seq_len);
}


//! Do simple operations work correctly?
BOOST_AUTO_TEST_CASE( sequence_filter )
{
  const char *core_seq = "AATTGGCC";
  Sequence s1(core_seq, reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(s1, core_seq);

  Sequence s2("aattggcc", reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(s2, "AATTGGCC");
  BOOST_CHECK_EQUAL(s2.rev_comp(), "GGCCAATT");
  BOOST_CHECK_EQUAL(s2.rev_comp(), "ggccaatt"); // we should be case insensitive now
  BOOST_CHECK_EQUAL(s2.size(), s2.size());
  //We're currently forcing sequences to uppercase
  BOOST_CHECK_EQUAL(s2.get_sequence(), "AATTGGCC"); 

  Sequence s3("asdfg", reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(s3, "ANNNG");
  BOOST_CHECK_EQUAL(s3.subseq(0,2), "AN");

  s3.set_filtered_sequence("AAGGCCTT", reduced_dna_alphabet, 0, 2); 
  BOOST_CHECK_EQUAL(s3, "AA");
  s3.set_filtered_sequence("AAGGCCTT", reduced_dna_alphabet, 2, 2);
  BOOST_CHECK_EQUAL( s3, "GG");
  s3.set_filtered_sequence("AAGGCCTT", reduced_dna_alphabet, 4);
  BOOST_CHECK_EQUAL( s3, "CCTT");

  s3 = "AAGGFF";
  BOOST_CHECK_EQUAL(s3, "AAGGNN");
}

BOOST_AUTO_TEST_CASE( sequence_nucleic_alphabet )
{
  std::string agct("AGCT");
  Sequence seq(agct, nucleic_alphabet);
  BOOST_CHECK_EQUAL(seq.size(), agct.size());
  BOOST_CHECK_EQUAL(seq.get_sequence(), agct);
  
  std::string bdv("BDv");
  Sequence seq_bdv(bdv, nucleic_alphabet);
  BOOST_CHECK_EQUAL(seq_bdv.size(), bdv.size());
  // forcing sequence to upper case
  BOOST_CHECK_EQUAL(seq_bdv.get_sequence(), 
                    boost::algorithm::to_upper_copy(bdv));
  
}

BOOST_AUTO_TEST_CASE( sequence_default_alphabet )
{
  std::string agct("AGCT");
  Sequence seq(agct);
  BOOST_CHECK_EQUAL(seq.size(), agct.size());
  BOOST_CHECK_EQUAL(seq.get_sequence(), agct);
  BOOST_CHECK_EQUAL(seq[0], agct[0]);
  BOOST_CHECK_EQUAL(seq[1], agct[1]);
  BOOST_CHECK_EQUAL(seq[2], agct[2]);
  BOOST_CHECK_EQUAL(seq[3], agct[3]);
  
  std::string bdv("BDv");
  Sequence seq_bdv(bdv);
  BOOST_CHECK_EQUAL(seq_bdv.size(), bdv.size());
  // default alphabet only allows AGCTUN
  BOOST_CHECK_EQUAL(seq_bdv.get_sequence(), "NNN");  
}

BOOST_AUTO_TEST_CASE( subseq_names )
{
  Sequence s1("AAGGCCTT", reduced_dna_alphabet);
  s1.set_species("species");
  s1.set_fasta_header("a fasta header");
  Sequence s2 = s1.subseq(2,2);
  BOOST_CHECK_EQUAL(s2, "GG");
  BOOST_CHECK_EQUAL(s2.get_species(), s1.get_species());
  BOOST_CHECK_EQUAL(s2.get_fasta_header(), s1.get_fasta_header());
}

BOOST_AUTO_TEST_CASE( sequence_start_stop )
{
  Sequence s1;
  BOOST_CHECK_EQUAL( s1.start(), 0 );
  BOOST_CHECK_EQUAL( s1.stop(), 0 );

  std::string seq_string("AAGGCCTT");
  Sequence s2(seq_string, reduced_dna_alphabet);
  BOOST_CHECK_EQUAL( s2.start(), 0 );
  BOOST_CHECK_EQUAL( s2.stop(), seq_string.size() );

  std::string s3seq_string = seq_string.substr(2,3);
  Sequence s3 = s2.subseq(2,3);
  BOOST_CHECK_EQUAL( s3.start(), 2);
  BOOST_CHECK_EQUAL( s3.stop(), 2+3);
  BOOST_CHECK_EQUAL( s3.size(), 3);
  BOOST_CHECK_EQUAL( s3, s3seq_string);
  
  std::string s4seq_string = s3seq_string.substr(1,1);
  Sequence s4 = s3.subseq(1,1);
  BOOST_CHECK_EQUAL( s4.start(), 1 );
  BOOST_CHECK_EQUAL( s4.stop(), 1+1);
  BOOST_CHECK_EQUAL( s4.size(), 1);
  BOOST_CHECK_EQUAL( s4, s4seq_string);
}

//! Can we load data from a file
BOOST_AUTO_TEST_CASE( sequence_load )
{
  fs::path seq_path(fs::path(EXAMPLE_DIR, fs::native)/ "seq");
  seq_path /=  "human_mck_pro.fa";
  Sequence s;
  s.load_fasta(seq_path, reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(s.subseq(0, 5), "GGATC"); // first few chars of fasta file
  BOOST_CHECK_EQUAL(s.subseq(2, 3), "ATC");
  BOOST_CHECK_EQUAL(s.get_fasta_header(), "gi|180579|gb|M21487.1|HUMCKMM1 Human "
                                    "muscle creatine kinase gene (CKMM), "
                                    "5' flank");
}

BOOST_AUTO_TEST_CASE( sequence_load_fasta_error )
{
  fs::path seq_path(fs::path(EXAMPLE_DIR, fs::native)/"seq");
  seq_path /= "broken.fa";
  bool exception_thrown = false;
  std::string exception_filename;
  Sequence s;
  try {
    s.load_fasta(seq_path);
  } catch(sequence_invalid_load_error e) {
    exception_thrown = true;
    size_t native_string_size = seq_path.native_file_string().size();
    std:string estr(e.what());
    BOOST_REQUIRE(estr.size() > native_string_size);
    std::copy(estr.begin(), estr.begin()+native_string_size,
              std::back_inserter(exception_filename));
  }
  BOOST_CHECK_EQUAL(exception_thrown, true);
  BOOST_CHECK_EQUAL(seq_path.native_file_string(), exception_filename);
}

BOOST_AUTO_TEST_CASE( sequence_load_annot_error )
{
  fs::path seq_path(fs::path(EXAMPLE_DIR, fs::native)/"seq");
  seq_path /= "mouse_mck_pro.fa";
  fs::path annot_path(fs::path(EXAMPLE_DIR, fs::native));
  annot_path /= "broken.annot";
  bool exception_thrown = false;
  Sequence s;
  s.load_fasta(seq_path);

  std::string exception_filename;
  try {
    s.load_annot(annot_path, 0, 0);
  } catch(annotation_load_error e) {
    exception_thrown = true;
    std:string estr(e.what());
    size_t native_string_size = annot_path.native_file_string().size();
    BOOST_REQUIRE(estr.size() > native_string_size);
    std::copy(estr.begin(), estr.begin()+native_string_size,
              std::back_inserter(exception_filename));
  }
  BOOST_CHECK_EQUAL(exception_thrown, true);
  BOOST_CHECK_EQUAL(annot_path.native_file_string(), exception_filename);
}

BOOST_AUTO_TEST_CASE( sequence_load_dna_reduced )
{
  std::string reduced_dna_fasta_string(">foo\nAAGGCCTTNN\n");
  std::stringstream reduced_dna_fasta(reduced_dna_fasta_string);
  std::string invalid_dna_fasta_string(">wrong\nAUSSI\n");
  std::stringstream invalid_dna_fasta(invalid_dna_fasta_string);
  std::string reduced_rna_fasta_string(">foo\nAAGGCCUUNN-\n");
  std::stringstream reduced_rna_fasta(reduced_rna_fasta_string);
  std::string garbage_string(">foo\na34ralk3547oilk,jual;(*&^#%-\n");
  std::stringstream garbage_fasta(garbage_string);
  
  Sequence s;
  s.load_fasta(reduced_dna_fasta, reduced_dna_alphabet);
  BOOST_CHECK_THROW(s.load_fasta(invalid_dna_fasta, 
                                 reduced_dna_alphabet),
                    sequence_invalid_load_error);
  BOOST_CHECK_THROW(s.load_fasta(reduced_rna_fasta, 
                                 reduced_dna_alphabet),
                    sequence_invalid_load_error);
  BOOST_CHECK_THROW(s.load_fasta(garbage_fasta, 
                                 reduced_dna_alphabet),
                    sequence_invalid_load_error);

}

BOOST_AUTO_TEST_CASE( sequence_load_rna_reduced )
{
  std::string reduced_rna_fasta_string(">foo\nAAGGCCUUNN\n");
  std::stringstream reduced_rna_fasta(reduced_rna_fasta_string);
  std::string invalid_rna_fasta_string(">wrong\nATSSI\n");
  std::stringstream invalid_rna_fasta(invalid_rna_fasta_string);
  std::string reduced_dna_fasta_string(">foo\nAAGGCCTTNN\n");
  std::stringstream reduced_dna_fasta(reduced_dna_fasta_string);
  std::string garbage_string(">foo\na34ralk3547oilk,jual;(*&^#%-\n");
  std::stringstream garbage_fasta(garbage_string);
  
  Sequence s;
  s.load_fasta(reduced_rna_fasta, reduced_rna_alphabet);
  BOOST_CHECK_THROW(s.load_fasta(invalid_rna_fasta, 
                                 reduced_rna_alphabet),
                    sequence_invalid_load_error);
  BOOST_CHECK_THROW(s.load_fasta(reduced_dna_fasta, 
                                 reduced_rna_alphabet),
                    sequence_invalid_load_error);
  BOOST_CHECK_THROW(s.load_fasta(garbage_fasta, 
                                 reduced_rna_alphabet),
                    sequence_invalid_load_error);
}

BOOST_AUTO_TEST_CASE( sequence_load_fasta_default )
{
  std::string reduced_rna_fasta_string(">foo\nAAGGCCUUNN\n");
  std::stringstream reduced_rna_fasta1(reduced_rna_fasta_string);
  std::stringstream reduced_rna_fasta2(reduced_rna_fasta_string);
  std::string invalid_nucleotide_fasta_string(">wrong\nATSSI\n");
  std::stringstream invalid_nucleotide_fasta(invalid_nucleotide_fasta_string);
  std::string reduced_dna_fasta_string(">foo\nAAGGCCTTNN\n");
  std::stringstream reduced_dna_fasta1(reduced_dna_fasta_string);
  std::stringstream reduced_dna_fasta2(reduced_dna_fasta_string);
  std::string garbage_string(">foo\na34ralk3547oilk,jual;(*&^#%-\n");
  std::stringstream garbage_fasta(garbage_string);
  
  Sequence s;
  Sequence specific;
  // there's two copies of reduced_rna_fasta because i didn't feel like
  // figuring out how to properly reset the read pointer in a stringstream
  s.load_fasta(reduced_rna_fasta1);
  specific.load_fasta(reduced_rna_fasta2, reduced_nucleic_alphabet);
  BOOST_CHECK_EQUAL(s, specific);
  
  s.load_fasta(reduced_dna_fasta1);
  specific.load_fasta(reduced_dna_fasta2, reduced_nucleic_alphabet);
  BOOST_CHECK_EQUAL(s, specific);
  
  BOOST_CHECK_THROW(s.load_fasta(invalid_nucleotide_fasta), 
                    sequence_invalid_load_error);
  BOOST_CHECK_THROW(s.load_fasta(garbage_fasta), 
                    sequence_invalid_load_error);
}

BOOST_AUTO_TEST_CASE( sequence_load_multiple_sequences_one_fasta ) 
{
  std::string fasta_file(
    ">gi|10129974|gb|AF188002.1|AF188002\n"
    "AAAAGGCTCCTGTCATATTGTGTCCTGCTCTGGTCTGC\n"
    ">gi|180579|gb|M21487.1|HUMCKMM1\n"
    "CGCCGAGAGCGCTTGCTCTGCCCAGATCTCGGCGAGTC\n"
    ">gi|1621|emb|X55146.1|OCMCK1\n"
    "CTCCCTGAGGGGAGTGCCCCGCTTAGCCC\n"
  );
  istringstream seq1_file(fasta_file);
  Sequence seq1;
  seq1.load_fasta(seq1_file, 1, 0, 0);
  BOOST_CHECK_EQUAL(seq1.get_sequence(), "AAAAGGCTCCTGTCATATTGTGTCCTGCTCTGGTCTGC");
  
  istringstream seq2_file(fasta_file);
  Sequence seq2;
  seq2.load_fasta(seq2_file, 2, 0, 0);
  BOOST_CHECK_EQUAL(seq2.get_sequence(), "CGCCGAGAGCGCTTGCTCTGCCCAGATCTCGGCGAGTC");
  
  istringstream seq3_file(fasta_file);
  Sequence seq3;
  seq3.load_fasta(seq3_file, 3, 0, 0);  
  BOOST_CHECK_EQUAL(seq3.get_sequence(), "CTCCCTGAGGGGAGTGCCCCGCTTAGCCC"); 
}

BOOST_AUTO_TEST_CASE( sequence_reverse_complement )
{
  std::string iupac_symbols("AaCcGgTtRrYySsWwKkMmBbDdHhVvNn-~.?");
  Sequence seq(iupac_symbols, nucleic_alphabet);
  Sequence seqr = seq.rev_comp();
  
  BOOST_CHECK( seq != seqr );
  BOOST_CHECK_EQUAL( seq, seqr.rev_comp() );
  // forcing sequence to upper case
  BOOST_CHECK_EQUAL( seq.get_sequence(), 
                     boost::algorithm::to_upper_copy(iupac_symbols) );
}

BOOST_AUTO_TEST_CASE( sequence_reverse_complement_dna )
{
  std::string dna_str("AGCTN");
  Sequence dna_seq(dna_str, reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(dna_seq.rev_comp(), "NAGCT");  
  BOOST_CHECK_EQUAL(dna_seq, dna_seq.rev_comp().rev_comp());
}

BOOST_AUTO_TEST_CASE( sequence_reverse_complement_rna )
{
  std::string rna_str("AGCUN");
  Sequence rna_seq(rna_str, reduced_rna_alphabet);
  BOOST_CHECK_EQUAL(rna_seq.rev_comp().get_sequence(), "NAGCU");  
  BOOST_CHECK_EQUAL(rna_seq.get_sequence(), 
                    rna_seq.rev_comp().rev_comp().get_sequence());
}

BOOST_AUTO_TEST_CASE( sequence_reverse_complement_subseq )
{
  std::string dna_str("AAAAAAAAAAGGGGGGGGGGG");
  Sequence seq(dna_str, reduced_dna_alphabet);
  Sequence subseq = seq.subseq(8,4);
  BOOST_CHECK_EQUAL( subseq, "AAGG");
  Sequence rev_subseq = subseq.rev_comp();
  BOOST_CHECK_EQUAL( rev_subseq.size(), subseq.size() );
  BOOST_CHECK_EQUAL( rev_subseq.get_sequence(), "CCTT");
}

BOOST_AUTO_TEST_CASE( sequence_reverse_iterator )
{
  std::string dna_str("AAAAAAAAAAGGGGGGGGGGG");
  std::string dna_str_reversed(dna_str.rbegin(), dna_str.rend());
  Sequence seq(dna_str);
  std::string seq_reversed(seq.rbegin(), seq.rend());
  BOOST_CHECK_EQUAL(seq_reversed, dna_str_reversed);
  
  std::string substr = dna_str.substr(8,4);
  Sequence subseq = seq.subseq(8,4);
  BOOST_CHECK_EQUAL(substr, subseq);

  std::string substr_reversed(substr.rbegin(), substr.rend());
  std::string subseq_reversed(subseq.rbegin(), subseq.rend());
  BOOST_CHECK_EQUAL(substr_reversed, subseq_reversed);  
}

BOOST_AUTO_TEST_CASE( sequence_empty_reverse_iterator)
{
  // so what happens with reverse interators when we have no sequence?
  Sequence seq1;
  Sequence seq2;
  Sequence seq3("AGCT");
  
  // all the empty sequences should have equal iterators
  BOOST_CHECK(seq1.rbegin() == seq1.rend());
  BOOST_CHECK(seq1.rbegin() == seq2.rend());
  
  // none of the seq1 iterators should equal any of the seq3 iterators
  BOOST_CHECK(seq1.rbegin() != seq3.rbegin());
  BOOST_CHECK(seq1.rbegin() != seq3.rend());
  BOOST_CHECK(seq1.rend() != seq3.rbegin());
  BOOST_CHECK(seq1.rend() != seq3.rend());
  
  // seq3 iterators should work
  BOOST_CHECK(seq3.rbegin() != seq3.rend());
  
}

BOOST_AUTO_TEST_CASE( annotation_load )
{
  string annot_data = "human\n"
                      "0 10 name   type\n"
                      "10 20 myf7\n"
                      "20 30 myod\n"
                      "50\t55 anothername\n"
                      "60 50 backward\n"
                      ">ident3 asdf\n"
                      "GCT\n"
                      "gCTn\n"
                      "75\t90\tname2\ttype2\n"
                      "100 120 name-asdf type!@#$%\n"
                      ;
  string s(100, 'A');
  s += "GCTGCTAATT";
  Sequence seq(s, reduced_dna_alphabet);
                     
  //istringstream annot_stream(annot_data);
  seq.parse_annot(annot_data, 0, 0);
  std::list<annot> annots_list = seq.annotations();
  std::vector<annot> annots(annots_list.begin(), annots_list.end());
  BOOST_REQUIRE_EQUAL( annots.size(), 8);
  BOOST_CHECK_EQUAL( annots[0].begin, 0 );
  BOOST_CHECK_EQUAL( annots[0].end, 10 );
  BOOST_CHECK_EQUAL( annots[0].type, "type");
  BOOST_CHECK_EQUAL( annots[0].name, "name");
  BOOST_CHECK_EQUAL( annots[1].name, "myf7");
  BOOST_CHECK_EQUAL( annots[2].name, "myod");
  BOOST_CHECK_EQUAL( annots[3].name, "anothername");
  BOOST_CHECK_EQUAL( annots[4].name, "backward");
  BOOST_CHECK_EQUAL( annots[5].name, "name2");
  BOOST_CHECK_EQUAL( annots[5].end, 90);
  BOOST_CHECK_EQUAL( annots[6].begin, 100);
  BOOST_CHECK_EQUAL( annots[6].end, 120);
  BOOST_CHECK_EQUAL( annots[6].name, "name-asdf");
  BOOST_CHECK_EQUAL( annots[6].type, "type!@#$%");
  // sequence defined annotations will always be after the
  // absolute positions
  BOOST_CHECK_EQUAL( annots[7].name, "ident3 asdf");
  BOOST_CHECK_EQUAL( annots[7].begin, 100);

  //BOOST_CHECK_EQUAL( annots
}

BOOST_AUTO_TEST_CASE( annotation_broken_load )
{
  string annot_data = "human\n"
                      "0 10 name   type\n"
                      "blah60 50 backward\n"
                      ">ident3 asdf\n"
                      "GCT\n"
                      "gCTn\n"
                      ;
  string s(100, 'A');
  s += "GCTGCTAATT";
  Sequence seq(s, reduced_dna_alphabet);
                     
  BOOST_CHECK_THROW(seq.parse_annot(annot_data, 0, 0), annotation_load_error);
  BOOST_CHECK_EQUAL(seq.annotations().size(), 0);
  }

BOOST_AUTO_TEST_CASE(annotation_ucsc_html_load)
{
  // this actually is basically what's returned by UCSC
  // (well actually with some of the sequence and copies of fasta blocks
  // removed to make the example shorter
  string annot_data = "\n"
    "<PRE>\n"
    ">hg17_knownGene_NM_001824_0 range=chr19:50517919-50517974 5'pad=0 3'pad=0 revComp=TRUE strand=- repeatMasking=none\n"
    "GGGTCAGTGTCACCTCCAGGATACAGACAG\n"
    "&gt;hg17_knownGene_NM_001824_3 range=chr19:50510563-50510695 5'pad=0 3'pad=0 revComp=TRUE strand=- repeatMasking=none\n"
    "GGTGGAGACGACCTGGACCCTAACTACGT\n"
    "</PRE>\n"
    "\n"
    "</BODY>\n"
    "</HTML>\n"
    ;

  string s = 
    "TGGGTCAGTGTCACCTCCAGGATACAGACAGCCCCCCTTCAGCCCAGCCCAGCCAG"
    "AAAAA"
    "GGTGGAGACGACCTGGACCCTAACTACGTGCTCAGCAGCCGCGTCCGCAC";
  Sequence seq(s, reduced_dna_alphabet);
  seq.parse_annot(annot_data);
  std::list<annot> annots = seq.annotations();
  BOOST_CHECK_EQUAL( annots.size(), 2);
}

BOOST_AUTO_TEST_CASE( annotation_load_no_species_name )
{
  string annot_data = "0 10 name   type\n"
                      "10 20 myf7\n"
                      "20 30 myod\n"
                      "50\t55 anothername\n"
                      "60 50 backward\n"
                      ">ident3 asdf\n"
                      "GCT\n"
                      "gCTn\n"
                      "75\t90\tname2\ttype2\n"
                      "100 120 name-asdf type!@#$%\n"
                      ;
  string s(100, 'A');
  s += "GCTGCTAATT";
  Sequence seq(s, reduced_dna_alphabet);
                     
  //istringstream annot_stream(annot_data);
  seq.parse_annot(annot_data, 0, 0);
  std::list<annot> annots_list = seq.annotations();
  std::vector<annot> annots(annots_list.begin(), annots_list.end());
  BOOST_REQUIRE_EQUAL( annots.size(), 8);
  BOOST_CHECK_EQUAL( annots[0].begin, 0 );
  BOOST_CHECK_EQUAL( annots[0].end, 10 );
  BOOST_CHECK_EQUAL( annots[0].type, "type");
}

// ticket:83 when you try to load a sequence from a file that doesn't
// have fasta headers it crashes. 
BOOST_AUTO_TEST_CASE( sequence_past_end ) 
{
  fs::path seq_path(fs::path(EXAMPLE_DIR, fs::native)/ "seq" );
  seq_path /=  "misformated_seq.fa";
  Sequence s;
  BOOST_CHECK_THROW( s.load_fasta(seq_path), mussa_load_error );
}

BOOST_AUTO_TEST_CASE ( sequence_empty )
{
  
  Sequence s;
  BOOST_CHECK_EQUAL( s.empty(), true );
  s = "AAAGGG";
  BOOST_CHECK_EQUAL( s.empty(), false );
  s.clear();
  BOOST_CHECK_EQUAL( s.empty(), true);
  s = "";
  BOOST_CHECK_EQUAL( s.empty(), true);
}

BOOST_AUTO_TEST_CASE ( sequence_size )
{
  
  Sequence s;
  BOOST_CHECK_EQUAL( s.size(), 0);
  std::string seq_string("AAAGGG");
  s = seq_string;
  BOOST_CHECK_EQUAL( s.size(), seq_string.size() );
  s.clear();
  BOOST_CHECK_EQUAL( s.size(), 0);
  s = "";
  BOOST_CHECK_EQUAL( s.size(), 0);
}

BOOST_AUTO_TEST_CASE( sequence_empty_equality )
{
  Sequence szero("", reduced_dna_alphabet);
  BOOST_CHECK_EQUAL(szero.empty(), true);
  BOOST_CHECK_EQUAL(szero, szero);
  BOOST_CHECK_EQUAL(szero, "");

  Sequence sclear("AGCT", reduced_dna_alphabet);
  sclear.clear();
  BOOST_CHECK_EQUAL(sclear.empty(), true);
  BOOST_CHECK_EQUAL(sclear, sclear);
  BOOST_CHECK_EQUAL(sclear, szero);
  BOOST_CHECK_EQUAL(sclear, "");

}
BOOST_AUTO_TEST_CASE ( sequence_iterators )
{
  std::string seq_string = "AAGGCCTTNNTATA";
  Sequence s(seq_string, reduced_dna_alphabet);
  const Sequence cs(s);
  std::string::size_type count = 0;

  std::string::iterator str_itor;
  Sequence::const_iterator s_itor;
  Sequence::const_iterator cs_itor;

  for( str_itor = seq_string.begin(),
       s_itor   = s.begin(),
       cs_itor  = cs.begin();
       str_itor != seq_string.end() and
       s_itor   != s.end() and
       cs_itor  != cs.end();
       ++str_itor, ++s_itor, ++cs_itor, ++count)
  {
    BOOST_CHECK_EQUAL ( *str_itor, *s_itor );
    BOOST_CHECK_EQUAL ( *s_itor, *cs_itor );
    BOOST_CHECK_EQUAL ( *cs_itor, *str_itor );
  }
  BOOST_CHECK_EQUAL( seq_string.size(), count );
  BOOST_CHECK_EQUAL( s.size(), count );
  BOOST_CHECK_EQUAL( cs.size(), count );
}

BOOST_AUTO_TEST_CASE( sequence_motifs )
{
  string m("AAAA");
  string bogus("AATTGGAA");
  Sequence s1("AAAAGGGGCCCCTTTT", reduced_dna_alphabet);

  list<motif>::const_iterator motif_i = s1.motifs().begin();
  list<motif>::const_iterator motif_end = s1.motifs().end();

  // do our iterators work?
  BOOST_CHECK( motif_i == s1.motifs().begin() );
  BOOST_CHECK( motif_end == s1.motifs().end() );
  BOOST_CHECK( motif_i == motif_end );

  // this shouldn't show up
  s1.add_motif(bogus);
  BOOST_CHECK( s1.motifs().begin() == s1.motifs().end() );
  BOOST_CHECK_EQUAL( s1.motifs().size(), 0 );

  s1.add_motif(m);
  BOOST_CHECK( s1.motifs().begin() != s1.motifs().end() );
  BOOST_CHECK_EQUAL( s1.motifs().size(), 2 );

  for(motif_i = s1.motifs().begin(); 
      motif_i != s1.motifs().end(); 
      ++motif_i)
  {
    BOOST_CHECK_EQUAL( motif_i->type, "motif" );
    BOOST_CHECK_EQUAL( motif_i->name, m);
    BOOST_CHECK_EQUAL( motif_i->sequence, m);
  }

  s1.clear_motifs();
  BOOST_CHECK( s1.motifs().begin() == s1.motifs().end() );

  /* FIXME: enable this when i find a way of passing storing the motif name
  // does our annotation travel?
  Sequence motif_seq(m);
  motif_seq.set_fasta_header("hi");
  s1.add_motif(motif_seq);

  BOOST_CHECK_EQUAL(s1.motifs().size(), 2);
  for(motif_i = s1.motifs().begin(); 
      motif_i != s1.motifs().end(); 
      ++motif_i)
  {
    BOOST_CHECK_EQUAL( motif_i->type, "motif" );
    BOOST_CHECK_EQUAL( motif_i->name, "hi");
    BOOST_CHECK_EQUAL( motif_i->sequence, m);
  }
  */
}

BOOST_AUTO_TEST_CASE( sequence_motif_subseq)
{
  // when searching for a motif on a subsequence we should 
  // only search the subsequence ticket:199
  string aaaa("AAAA");
  string cccc("CCCC");
  Sequence s1("AAAANCCCC", reduced_dna_alphabet);

  // this shouldn't show up
  s1.add_motif(cccc);
  BOOST_CHECK_EQUAL( s1.motifs().size(), 1 );

  s1.add_motif(aaaa);
  BOOST_CHECK_EQUAL( s1.motifs().size(), 2 );

  Sequence subseq1 = s1.subseq(4,5);
  BOOST_CHECK_EQUAL(subseq1.motifs().size(), 2);
  subseq1.clear_motifs();
  BOOST_CHECK_EQUAL(subseq1.motifs().size(), 0);
  // this is outside of our subsequence, and so shouldn't be found    
  subseq1.add_motif(aaaa);
  BOOST_CHECK_EQUAL( subseq1.motifs().size(), 0 );
  
  subseq1.add_motif(cccc);
  BOOST_CHECK_EQUAL( subseq1.motifs().size(), 1);
  std::list<motif>::const_iterator motif_i = subseq1.motifs().begin();
  BOOST_REQUIRE(motif_i != subseq1.motifs().end());
  BOOST_CHECK_EQUAL(motif_i->begin, 1);
  BOOST_CHECK_EQUAL(motif_i->end, 5);
}

BOOST_AUTO_TEST_CASE( annot_test )
{
  annot a(0, 10, "test", "thing");

  BOOST_CHECK_EQUAL( a.begin, 0 );
  BOOST_CHECK_EQUAL( a.end,   10 );
  BOOST_CHECK_EQUAL( a.type,  "test" );
  BOOST_CHECK_EQUAL( a.name,  "thing" );

  motif m(10, "AAGGCC");
  BOOST_CHECK_EQUAL( m.begin, 10 );
  BOOST_CHECK_EQUAL( m.type, "motif" );
  BOOST_CHECK_EQUAL( m.name, "AAGGCC" );
  BOOST_CHECK_EQUAL( m.end,  10+6 );
}

BOOST_AUTO_TEST_CASE( annotate_from_sequence )
{
  string s("CCGCCCCCCATCATCGCGGCTCTCCGAGAGTCCCGCGCCCCACTCCCGGC"
           "ACCCACCTGACCGCGGGCGGCTCCGGCCCCGCTTCGCCCCACTGCGATCA"
           "GTCGCGTCCCGCAGGCCAGGCACGCCCCGCCGCTCCCGCTGCGCCGGGCG"
           "TCTGGGACCTCGGGCGGCTCCTCCGAGGGGCGGGGCAGCCGGGAGCCACG"
           "CCCCCGCAGGTGAGCCGGCCACGCCCACCGCCCGTGGGAAGTTCAGCCTC"
           "GGGGCTCCAGCCCCGCGGGAATGGCAGAACTTCGCACGCGGAACTGGTAA"
           "CCTCCAGGACACCTCGAATCAGGGTGATTGTAGCGCAGGGGCCTTGGCCA"
           "AGCTAAAACTTTGGAAACTTTAGATCCCAGACAGGTGGCTTTCTTGCAGT");
  string gc("GCCCCC");
  string gga("GGACACCTC");
  Sequence seq(s, reduced_dna_alphabet);

  std::list<Sequence> query_list;
  std::list<string> string_list;
  query_list.push_back(Sequence(gc));
  string_list.push_back(gc);
  query_list.push_back(Sequence(gga));
  string_list.push_back(gga);

  BOOST_CHECK_EQUAL( seq.annotations().size(), 0 );
  seq.find_sequences(query_list.begin(), query_list.end());
  
  int count = 0;
  for(list<string>::iterator string_i = string_list.begin();
      string_i != string_list.end();
      ++string_i)
  {
    string::size_type pos=0;
    while(pos != string::npos) {
      pos = s.find(*string_i, pos);
      if (pos != string::npos) {
        ++count;
        ++pos;
      }
    }
  }
  BOOST_CHECK_EQUAL(seq.annotations().size(), count);
  const std::list<annot> &a = seq.annotations();
  for (std::list<annot>::const_iterator annot_i = a.begin();
       annot_i != a.end();
       ++annot_i)
  {
    int count = annot_i->end - annot_i->begin ;
  }
}

BOOST_AUTO_TEST_CASE( subseq_annotation_test )
{
  string s("CCGCCCCCCATCATCGCGGCTCTCCGAGAGTCCCGCGCCCCACTCCCGGC"
           "ACCCACCTGACCGCGGGCGGCTCCGGCCCCGCTTCGCCCCACTGCGATCA"
           "GTCGCGTCCCGCAGGCCAGGCACGCCCCGCCGCTCCCGCTGCGCCGGGCG"
           "TCTGGGACCTCGGGCGGCTCCTCCGAGGGGCGGGGCAGCCGGGAGCCACG"
           "CCCCCGCAGGTGAGCCGGCCACGCCCACCGCCCGTGGGAAGTTCAGCCTC"
           "GGGGCTCCAGCCCCGCGGGAATGGCAGAACTTCGCACGCGGAACTGGTAA"
           "CCTCCAGGACACCTCGAATCAGGGTGATTGTAGCGCAGGGGCCTTGGCCA"
           "AGCTAAAACTTTGGAAACTTTAGATCCCAGACAGGTGGCTTTCTTGCAGT");
  Sequence seq(s, reduced_dna_alphabet);


  seq.add_annotation(annot(0, 10, "0-10", "0-10"));
  seq.add_annotation(annot(10, 20, "10-20", "10-20"));
  seq.add_annotation(annot(0, 20, "0-20", "0-20"));
  seq.add_annotation(annot(8, 12, "8-12", "8-12"));
  seq.add_annotation(annot(100, 5000, "100-5000", "100-5000"));

  Sequence subseq = seq.subseq(5, 10);
  const list<annot> annots = subseq.annotations();
  // generate some ground truth
  list<annot> correct;
  correct.push_back(annot(0, 5, "0-10",  "0-10"));
  correct.push_back(annot(5,10, "10-20", "10-20"));
  correct.push_back(annot(0,10, "0-20",  "0-20"));
  correct.push_back(annot(3, 7, "8-12",  "8-12"));
  BOOST_REQUIRE_EQUAL( annots.size(), correct.size() );

  list<annot>::iterator correct_i = correct.begin();
  list<annot>::const_iterator annot_i = annots.begin();
  for(; annot_i != annots.end(); ++annot_i, ++correct_i)
  {
    BOOST_CHECK( *annot_i == *correct_i );
  }
}

BOOST_AUTO_TEST_CASE( motif_annotation_update )
{
  string s("CCGTCCCCCATCATCGCGGCTCTCCGAGAGTCCCGCGCCCCACTCCCGGC"
           "ACCCACCTGACCGCGGGCGGCTCCGGCCCCGCTTCGCCCCACTGCGATCA"
           "GTCGCGTCCCGCAGGCCAGGCACGCCCCGCCGCTCCCGCTGCGCCGGGCG"
           "TCTGGGACCTCGGGCGGCTCCTCCGAGGGGCGGGGCAGCCGGGAGCCACG"
           "CCCCCGCAGGTGAGCCGGCCACGCCCACCGCCCGTGGGAAGTTCAGCCTC"
           "GGGGCTCCAGCCCCGCGGGAATGGCAGAACTTCGCACGCGGAACTGGTAA"
           "CCTCCAGGACACCTCGAATCAGGGTGATTGTAGCGCAGGGGCCTTGGCCA"
           "AGCTAAAACTTTGGAAACTTTAGATCCCAGACAGGTGGCTTTCTTGCAGT");
  Sequence seq(s, reduced_dna_alphabet);

  // starting conditions
  BOOST_CHECK_EQUAL(seq.annotations().size(), 0);
  BOOST_CHECK_EQUAL(seq.motifs().size(), 0);
  seq.add_annotation(annot(0, 10, "0-10", "0-10"));
  seq.add_annotation(annot(10, 20, "10-20", "10-20"));
  seq.add_annotation(annot(0, 20, "0-20", "0-20"));
  BOOST_CHECK_EQUAL(seq.annotations().size(), 3);
  BOOST_CHECK_EQUAL(seq.motifs().size(), 0);
  seq.add_motif("CCGTCCC");
  BOOST_CHECK_EQUAL(seq.annotations().size(), 3);
  BOOST_CHECK_EQUAL(seq.motifs().size(), 1);
  seq.clear_motifs();
  BOOST_CHECK_EQUAL(seq.annotations().size(), 3);
  BOOST_CHECK_EQUAL(seq.motifs().size(), 0);
}

BOOST_AUTO_TEST_CASE( out_operator )
{
  string s("AAGGCCTT");
  Sequence seq(s, reduced_dna_alphabet);

  ostringstream buf;
  buf << s;
  BOOST_CHECK_EQUAL( s, buf.str() );
}

BOOST_AUTO_TEST_CASE( get_name )
{
  Sequence seq("AAGGCCTT", reduced_dna_alphabet);

  BOOST_CHECK_EQUAL( seq.get_name(), "" );
  seq.set_species("hooman"); // anyone remember tradewars?
  BOOST_CHECK_EQUAL( seq.get_name(), "hooman");
  seq.set_fasta_header("fasta human");
  BOOST_CHECK_EQUAL( seq.get_name(), "fasta human");
}

BOOST_AUTO_TEST_CASE( serialize_simple )
{
  std::string seq_string = "AAGGCCTT";
  Sequence seq(seq_string, reduced_dna_alphabet);
  seq.set_species("ribbet");
  std::ostringstream oss;
  // allocate/deallocate serialization components
  {
    boost::archive::text_oarchive oarchive(oss);
    const Sequence& const_seq(seq);
    BOOST_CHECK_EQUAL(seq, const_seq);
    oarchive << const_seq;
  }
  Sequence seq_loaded;
  {
    std::istringstream iss(oss.str());
    boost::archive::text_iarchive iarchive(iss);
    iarchive >> seq_loaded;
  }
  BOOST_CHECK_EQUAL(seq_loaded, seq);
  BOOST_CHECK_EQUAL(seq.get_species(), "ribbet");
}  

BOOST_AUTO_TEST_CASE( serialize_tree )
{
  std::string seq_string = "AAGGCCTT";
  Sequence seq(seq_string, reduced_dna_alphabet);
  seq.set_species("ribbet");
  seq.add_motif("AA");
  seq.add_motif("GC");
  annot a1(6,7,"t","t");
  seq.add_annotation(a1);

  std::ostringstream oss;
  // allocate/deallocate serialization components
  {
    boost::archive::text_oarchive oarchive(oss);
    const Sequence& const_seq(seq);
    BOOST_CHECK_EQUAL(seq, const_seq);
    oarchive << const_seq;
  }

  Sequence seq_loaded;
  {
    std::istringstream iss(oss.str());
    boost::archive::text_iarchive iarchive(iss);
    iarchive >> seq_loaded;
  }
  BOOST_CHECK_EQUAL(seq_loaded, seq);
}  

// this writes out an "old" style annotated sequence
// with annotations attached as "motifs" and "annots"
BOOST_AUTO_TEST_CASE( serialize_xml_sequence )
{
  std::string seq_string = "AAGGCCTT";
  Sequence seq(seq_string, reduced_dna_alphabet);
  seq.set_species("ribbet");
  seq.add_motif("AA");
  seq.add_motif("GC");
  annot a1(6,7,"t","t");
  seq.add_annotation(a1);

  std::ostringstream oss;
  // allocate/deallocate serialization components
  {
    boost::archive::xml_oarchive oarchive(oss);
    const Sequence& const_seq(seq);
    BOOST_CHECK_EQUAL(seq, const_seq);
    oarchive << boost::serialization::make_nvp("root", const_seq);
  }
  Sequence seq_loaded;
  {
    std::istringstream iss(oss.str());
    boost::archive::xml_iarchive iarchive(iss);
    iarchive >> boost::serialization::make_nvp("root", seq_loaded);
  }
  BOOST_CHECK_EQUAL(seq_loaded, seq);
}

BOOST_AUTO_TEST_CASE( serialize_xml_two )
{
  std::string seq_string = "AAGGCCTT";
  Sequence seq1(seq_string, reduced_dna_alphabet);
  Sequence seq2(seq1);

  std::ostringstream oss;
  // allocate/deallocate serialization components
  {
    boost::archive::xml_oarchive oarchive(oss);
    const Sequence& const_seq1(seq1);
    const Sequence& const_seq2(seq2);
    oarchive << boost::serialization::make_nvp("seq1", const_seq1);
    oarchive << boost::serialization::make_nvp("seq2", const_seq2);
  }
  //std::cout << "xml: " << oss.str() << std::endl;
  Sequence seq1_loaded;
  Sequence seq2_loaded;
  {
    std::istringstream iss(oss.str());
    boost::archive::xml_iarchive iarchive(iss);
    iarchive >> boost::serialization::make_nvp("seq1", seq1_loaded);
    iarchive >> boost::serialization::make_nvp("seq2", seq2_loaded);
  }
  BOOST_CHECK_EQUAL(seq1_loaded, seq1);
  BOOST_CHECK_EQUAL(seq2_loaded, seq2);
  // test if our pointers are the same
  BOOST_CHECK_EQUAL(seq1_loaded.data(), seq2_loaded.data());
}
